/* $Id: fbcon-kgi.c,v 1.77 2000/08/24 07:47:26 marcus Exp $
******************************************************************************

   KGI-fbdev bridge layer

   Copyright (C) 1998      Andrew Apted		[andrew@ggi-project.org]
   Copyright (C) 1998-1999 Jon Taylor		[taylorj@ggi-project.org]
   Copyright (C) 1998-1999 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

/*
 * Uncomment to allow non-root users to map the MMIO region.
#define NONROOT_MMIO   1
*/

#undef KGICON_USE_PROC	/* Doesn't work yet */

#define KGICON_USE_MMAP 1
#define USE_VMOPS       1

#define DEBUG_LEVEL	0

#include <kgi/maintainers.h>
#define MAINTAINER	Jon_Taylor
#define DRIVER_NAME	"KGI-fbdev bridge layer"
#define DRIVER_REV	"$Revision: 1.77 $"

/* This need to come first for modversions to work */
#include <kgi/module.h>

#include <linux/fb.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

struct kgi_driver fbcon_driver_id={
	"GGI project members.",
	DRIVER_NAME,
	0x00800001		/* KGI 0.80, fbcon bridge version 0.01 */
};


#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#include <video/fbcon-vga.h>
#include <video/font.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
# ifndef LINUX_23
#  define LINUX_23 1
# endif
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
# ifndef LINUX_24
#  define LINUX_24 1
# endif
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,25)
# define HAVE_VMA_VM_PGOFF 1
#endif
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,99))
# define HAVE_STRUCT_VM_OPERATIONS_ADVISE 1
#endif

#define KGICON_IO_BUF_SIZE     2*PAGE_SIZE


/* Param/Sysctl data
 */
#include <linux/sysctl.h>
#include <linux/proc_fs.h>

#define SYSCTL_KGICON_BASE 1147	/* Any chance we can get a range of
				   numbers allocated from Linus? */
#define KGI_SUBSYS_LEAFS 0

static ctl_table sysctl_subsys[KGI_NR_SUBSYSTEMS+KGI_SUBSYS_LEAFS+1] = {
	/*	
	   ...
	   Add generic leafs here (don't forget to change KGI_SUBSYS_LEAFS)
	   ...
	*/
	{0}
};

static char sysctl_fb_string[12];

static ctl_table sysctl_fbnum[] = {
	{0, sysctl_fb_string, NULL, 0, 0555, sysctl_subsys},
	{0, }
};

static struct ctl_table_header *sysctl_th = NULL;

static void kgisysctl_init(int fbidx);
static void kgisysctl_exit(void);
#if defined(CONFIG_PROC_FS) && defined(KGICON_USE_PROC)
static int  kgiproc_init(void);
static void kgiproc_exit(void);
static void
pingpong_exec(struct kgi_accel *accel, struct kgi_accel_buffer *peng);
static void pingpong_wait(struct kgi_accel *accel);
static struct kgi_accel_mapping accel_map;
#else
#undef KGICON_USE_PROC
#define kgiproc_init()	(0)
#define kgiproc_exit()	/* no-op */
#endif


/* Internal info structure
 */
struct kgiproc_entry {
	struct proc_dir_entry *entry;
	struct kgiproc_entry  *next;
};

struct fb_info_kgicon
{
	struct fb_info info;
	struct fb_fix_screeninfo fix;

	struct fb_var_screeninfo default_var;
	struct display default_display;

	kgi_color palette[256];

	unsigned long videophys;
	unsigned long videovirt;
	unsigned long videosize;

	struct kgiproc_entry *proc_entry;
	int mtrr_entry;

#ifdef FBCON_HAS_CFB16
	__u16 cmap_cfb15[16];
	__u16 cmap_cfb16[16];
#endif
#ifdef FBCON_HAS_CFB24
	__u32 cmap_cfb24[16];
#endif
#ifdef FBCON_HAS_CFB32
	__u32 cmap_cfb32[16];
#endif
	/* Used for kgi mmio acceleration */
	unsigned long mmio_start;
	unsigned long mmio_len;
	unsigned long accel;
};

static struct fb_info_kgicon the_info;

/* Parameters
 */
int kgicon_width   = 640;	/* initial geometry */
int kgicon_height  = 480;
int kgicon_bpp     = 1;		/* <- should be 8, but VGA driver is buggy */
int kgicon_inverse = 0;
int ramdac_mhz     = 0;
int kgicon_mtrr    = 1;

#ifdef MODULE_PARM
MODULE_PARM(kgicon_width,   "1-2048i");
MODULE_PARM(kgicon_height,  "1-2048i");
MODULE_PARM(kgicon_bpp,     "1-32i");
MODULE_PARM(kgicon_inverse, "0-1i");
MODULE_PARM(ramdac_mhz,     "1-1024i");
MODULE_PARM(kgicon_mtrr,    "0-1i");
#endif


/* Interface used by the world
*/
void kgifb_setup(char *options, int *ints);
static int kgifb_open(struct fb_info *info, int user);
static int kgifb_release(struct fb_info *info, int user);
static int kgifb_get_fix(struct fb_fix_screeninfo *fix, int con,
			 struct fb_info *info);
static int kgifb_get_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info);
static int kgifb_set_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info);
static int kgifb_pan_display(struct fb_var_screeninfo *var, int con,
			     struct fb_info *info);
static int kgifb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			  struct fb_info *info);
static int kgifb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);
static int kgifb_ioctl(struct inode *inode, struct file *file, u_int cmd,
		       u_long arg, int con, struct fb_info *info);
#ifdef KGICON_USE_MMAP
static int kgifb_mmap(struct fb_info *info, struct file *file,
		      struct vm_area_struct *vma);
#endif

/* Interface to the low level console driver
*/
int kgifb_init(void);
static int kgifb_con_switch(int con, struct fb_info *info);
static int kgifb_con_updatevar(int con, struct fb_info *info);
static void kgifb_con_blank(int blank, struct fb_info *info);

/* Internal routines
*/
static u_long get_stride(int xres_virtual, int bpp);
static void kgifb_encode_fix(struct fb_fix_screeninfo *fix,
			     struct fb_var_screeninfo *var);
static void set_color_bitfields(struct fb_var_screeninfo *var);
static int kgifb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
			   u_int *transp, struct fb_info *info);
static int kgifb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
			   u_int transp, struct fb_info *info);
static void do_install_cmap(int con, struct fb_info *info);

/* KGI functions
*/
void kgi_set_origin(struct kgi_device *dev, kgi_uint frame, 
		    kgi_uint x, kgi_uint y)
{ 
}

void kgi_set_split(struct kgi_device *dev, kgi_uint line)
{ 
}

struct kgi_font *kgi_default_font(kgi_modereq *mode) 
{ 
	return NULL; 
}

int kgi_register_display(struct kgi_display *dpy, int id)
{ 
	return 0; 
}

kgi_fb_ops text16_color_io  = 
{ 
};

kgi_fb_ops text16_color_mem = 
{ 
};

static struct kgi_font kgicon_font =
{
	8, 16, 8, 16, 11,
	NULL, NULL,
};

/* This stuff should be picked up from the KGI driver */
#define	CONFIG_CARD_MANUFACT	"Manufacturer"
#define	CONFIG_CARD_MODEL	"Model"
#define	CONFIG_CARD_MODES	GM_ALL
#define	CONFIG_CARD_MAX_X	0x7FFF
#define	CONFIG_CARD_MAX_Y	0x7FFF

static struct kgi_card kgifb_card = 
{
	CONFIG_CARD_MANUFACT,			/* Card manufacturer	*/
	CONFIG_CARD_MODEL,			/* Card model		*/
	CONFIG_CARD_MODES,			/* Allowed modes	*/
	0,
	{ CONFIG_CARD_MAX_X, CONFIG_CARD_MAX_Y },/* User limits		*/
	NULL,	/* driver private data hook for card as a whole		*/
	NULL,	
	NULL,	/* card component info structs, these are not known yet	*/
	NULL,	/* and get filled in during inititalization		*/
	NULL,
	{	/* The mode to set when resetting the card		*/
		{ 1,				/* need one frame	*/
		  { 80, 25 }, { 80, 25 },	/* visible/virtual	*/
		  { 0, 0 },			/* size in mm		*/
		  KGIGT_TEXT16,			/* mode			*/
		  { 9, 16 }			/* text size		*/
		},
		{ 720, 720, 765, 873, 882, 900, 0 },	/* horiz.timing	*/
		{ 400, 407, 412, 414, 442, 449, 1 },	/* vert. timing	*/
		{ 1, 1 },			/* no magnification	*/
		0,				/* special capabilities	*/
		28322000,			/* dot clock in Hz	*/
		{ 1, 1 },			/* clock rate to program*/
		{ 1, 1 },			/* transfer rate to dac	*/

		0,
        }
};

static int kgim_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
			   enum kgi_tm_cmd cmd)
{
	int cnt = 10;
	int err;

	if (cmd == CMD_PROPOSE) {
		kgi_modereq request = mode->request;

		memset(mode, 0, sizeof(struct kgi_mode));

		mode->request = request;
	}

	while (cnt-- && !(cmd == CMD_READY)) {
		DEBUG2("check_mode:cmd=%i; cnt=%i", cmd, cnt);

		err = kgim_clock_check_mode (dpy, mode, &cmd);
		if (err) return err;
		DEBUG2("check_mode: clock OK");

		err = kgim_ramdac_check_mode(dpy, mode, &cmd);
		if (err) return err;
		DEBUG2("check_mode: ramdac OK");

		err = kgim_chipset_check_mode(dpy, mode, &cmd);
		if (err) return err;
		DEBUG2("check_mode: chipset OK");

		if (cmd == CMD_LOWER || cmd == CMD_RAISE) {
			err = kgim_clock_check_mode(dpy, mode, &cmd);
			if (err) return err;
			DEBUG2("Second clock check OK");
		}

		err = kgim_accel_check_mode(dpy, mode, &cmd);
		if (err) return err;
		DEBUG2("check_mode: accel OK");
 
		err = kgim_monitor_check_mode(dpy, mode, &cmd);
		if (err) return err;
		DEBUG2("check_mode: monitor OK");
	}
 
	return (++cnt) ? 0 : -EINVAL;
}


static void kgim_reset(struct kgi_display *dpy)
{
	kgim_chipset_set_mode(dpy, &(dpy->card->preset));

	/*	frame and origin are set to zero and splitline is set to
	**	dpy->card->reset->request.visible.y by the chipset driver
	**	during dpy->set_mode. Thus no need to reset them again.
	*/
}

static struct kgim_options_pci options_pci;
static struct kgim_options_misc options_misc;
static struct kgim_options_chipset options_chipset;

static struct kgim_options options =
{
	&options_misc,
	&options_pci,
	&options_chipset
};

static struct kgim_private module_private =
{
	&options,
	NULL
};

static kgi_modereq kgim_default_mode;

static struct kgi_graphic kgifb_graph;

static struct kgi_display kgifb_display = 
{
	-1,				/* id number			*/
	0,				/* refcnt			*/
	0xFFFFFFFF,			/* conflict_text		*/
	0xFFFFFFFF,			/* conflict_graph		*/
	0xFFFFFFFF,			/* conflict_setup		*/
	NULL,				/* currently mapped device	*/
	&module_private,		/* driver_data			*/
					/* these are set by the driver	*/
	NULL,				/* enable(struct kgi_display *)	*/
	NULL,				/* disable(struct kgi_display *)*/
	kgim_reset,			/* reset(struct kgi_display *)	*/
	{				/* ioctl methods		*/
		NULL,
		kgim_monitor_command,
		kgim_ramdac_command,
		kgim_clock_command,
		kgim_chipset_command,
		kgim_accel_command,
		kgim_vidio_command,
		NULL
	},
	&kgifb_card,			/* hardware info		*/
	kgim_check_mode,		/* check_mode()			*/
	kgim_chipset_set_mode,		/* set_mode()	-> chipset	*/
	NULL,				/* set_font()	-> chipset	*/
	NULL,				/* set_clut()	-> dac		*/
	NULL,				/* set_split()	-> chipset	*/
	NULL,				/* set_frame()	-> chipset	*/
	NULL,				/* set_origin()	-> chipset	*/
	NULL,				/* set_pointer() -> dac/chipset	*/
	NULL, NULL, NULL,		/* { show, hide, undo }_mouse	*/
	NULL, NULL, NULL,		/* { show, hide, undo }_cursor	*/
	{ {0}, },			/* current mode			*/
	&kgim_default_mode,		/* default mode of display	*/
	{ 0, }				/* framebuffer of current frame	*/
};

/* This is only called after io_buf and gc have been successfully allocated */
static void kgi_freemem(void)
{
#ifdef KGICON_USE_PROC
	free_page((unsigned long)accel_map.ping.buffer);
	free_page((unsigned long)accel_map.pong.buffer);
	free_page((unsigned long)accel_map.accel->name);
	free_page((unsigned long)accel_map.accel);
#endif
	vfree(kgifb_graph.io_buf);
	free_page((unsigned long)kgifb_graph.gc);
}

static int init_kgi(void)
{
	/*	NOTE: The various *_init() functions have to printk() the
	**	reason for failing if they fail.
	*/
	int err;

	the_info.mtrr_entry = -1;

	memset(&options_pci,  0, sizeof(options_pci));
	memset(&options_misc, 0, sizeof(options_misc));
	memset(&kgifb_graph,  0, sizeof(kgifb_graph));
#ifdef KGICON_USE_PROC
	memset(&accel_map,    0, sizeof(accel_map));
#endif

	options_pci.dev = PCICFG_NULL;

	/*      Setup kgifb_graph
	 */
	if ((kgifb_graph.gc = (kgi_gc *) get_free_page(GFP_ATOMIC)) == NULL) {
		return -ENOMEM;
	}

	if ((kgifb_graph.io_buf = vmalloc(KGICON_IO_BUF_SIZE)) == NULL) {
		free_page((unsigned long)kgifb_graph.gc);
		return -ENOMEM;
	}

#ifdef KGICON_USE_PROC	
	if ((accel_map.accel = (struct kgi_accel *) get_free_page(GFP_ATOMIC)) == NULL)
	{
		vfree(kgifb_graph.io_buf);
		free_page((unsigned long)kgifb_graph.gc);
		return -ENOMEM;
	}
	
	if ((accel_map.accel->name = (char *) get_free_page(GFP_ATOMIC)) == NULL)
	{
		free_page((unsigned long)accel_map.accel);
		vfree(kgifb_graph.io_buf);
		free_page((unsigned long)kgifb_graph.gc);
		return -ENOMEM;
	}
	
	if ((accel_map.ping.buffer = (void *) get_free_page(GFP_ATOMIC)) == NULL)
	{
		free_page((unsigned long)accel_map.accel->name);
		free_page((unsigned long)accel_map.accel);
		vfree(kgifb_graph.io_buf);
		free_page((unsigned long)kgifb_graph.gc);
		return -ENOMEM;
	}
	
	if ((accel_map.pong.buffer = (void *) get_free_page(GFP_ATOMIC)) == NULL)
	{
		free_page((unsigned long)accel_map.ping.buffer);
		free_page((unsigned long)accel_map.accel->name);
		free_page((unsigned long)accel_map.accel);
		vfree(kgifb_graph.io_buf);
		free_page((unsigned long)kgifb_graph.gc);
		return -ENOMEM;
	}
	
	sprintf(accel_map.accel->name, "kgicommand");
	
	memset(accel_map.ping.buffer, 0, PAGE_SIZE);
	memset(accel_map.pong.buffer, 0xab, PAGE_SIZE);
	
	accel_map.ping.size = accel_map.pong.size = PAGE_SIZE;
	accel_map.devid = 0;
	accel_map.accel->execute = pingpong_exec;
	accel_map.accel->wait = pingpong_wait;
#endif /* KGICON_USE_PROC */
	
	kgifb_graph.io_cnt = 0;
	kgifb_graph.io_lock = 0;
	kgifb_graph.dev.dpy = &kgifb_display;
	
	/*	The chipset provides I/O functions for all other parts, so we
	**	have to init it first.
	*/
	if ((err = kgim_chipset_init(&kgifb_card, &kgifb_display))) 
	{
		ERROR("kgim_chipset_init failed");
		kgi_freemem();
		return err;
	}

	/*	Now that the chipset is OK, we try to init the monitor driver.
	**	if this fails, we have to deinitialize the chipset.
	*/
	if ((err = kgim_monitor_init(&kgifb_card, &kgifb_display))) 
	{
		ERROR("kgim_monitor_init failed");
		kgim_chipset_done(&kgifb_display);
		kgi_freemem();
		return err;
	}

	/*	Now we go for the DAC. If it fails, we have to deinitialize
	**	the subsystems already fired up.
	*/
	if ((err = kgim_ramdac_init(&kgifb_card, &kgifb_display))) 
	{
		ERROR("kgim_ramdac_init failed");

		kgim_monitor_done(&kgifb_display);
		kgim_chipset_done(&kgifb_display);
		kgi_freemem();
		return -EIO;
	}

	/*	Now the clock driver. If this fails, ...
	*/
	if ((err = kgim_clock_init(&kgifb_card, &kgifb_display))) 
	{
		ERROR("kgim_clock_init failed");

		kgim_ramdac_done(&kgifb_display);
		kgim_monitor_done(&kgifb_display);
		kgim_chipset_done(&kgifb_display);
		kgi_freemem();
		return -EIO;
	}

	/*	Last is the accel driver. If this fails, ...
	*/
	if ((err = kgim_accel_init(&kgifb_card, &kgifb_display))) 
	{
		ERROR("kgim_accel_init failed");

		kgim_clock_done(&kgifb_display);
		kgim_ramdac_done(&kgifb_display);
		kgim_monitor_done(&kgifb_display);
		kgim_chipset_done(&kgifb_display);
		kgi_freemem();
		return -EIO;
	}

	/*	Last is the videoio driver. If this fails, ...
	*/
	if ((err = kgim_vidio_init(&kgifb_card, &kgifb_display))) 
	{
		ERROR("kgim_vidio_init failed");

		kgim_accel_done(&kgifb_display);
		kgim_clock_done(&kgifb_display);
		kgim_ramdac_done(&kgifb_display);
		kgim_monitor_done(&kgifb_display);
		kgim_chipset_done(&kgifb_display);
		kgi_freemem();
		return -EIO;
	}

	/*	Now set the initial display mode and other state. We know
	**	the mode check can't fail, because if it does we should never
	**	have gotten here.
	*/
	
	memcpy(kgifb_display.defmode, &kgifb_card.preset.request, 
	       sizeof(kgi_modereq));
	memcpy(&kgifb_display.mode.request, kgifb_display.defmode, 
	       sizeof(kgi_modereq));

	NOTICE("%s %s driver loaded.",
	       kgifb_card.manufact, kgifb_card.model);

	return 0;
}


static void exit_kgi(void)
{
	/* Restore card into the mode in was in beforehand.
	 */

#ifdef CONFIG_MTRR
	if (the_info.mtrr_entry >= 0) {
		mtrr_del(the_info.mtrr_entry, 0, 0);
	}
#endif
        kgim_reset(&kgifb_display);

        kgim_vidio_done(&kgifb_display);
        kgim_accel_done(&kgifb_display);
        kgim_clock_done(&kgifb_display);
        kgim_ramdac_done(&kgifb_display);
        kgim_monitor_done(&kgifb_display);
        kgim_chipset_done(&kgifb_display);
	kgi_freemem();
	NOTICE("exit_kgi: Finished.");
}


/* ==================================================================== */

static int currcon = 0;


typedef struct kgicon_defmode
{
	int width, height, bpp;	   /* 0 means use `kgicon_XXX' */
} kgicon_defmode;


#define NR_DEFAULT_MODES  30

static kgicon_defmode default_modes[NR_DEFAULT_MODES] =
{
	{ 0, 0, 0 },	/* first try command-line parameters */

	{ 0, 0, 1 },
	{ 0, 0, 8 },
	{ 0, 0, 16 },
	{ 0, 0, 24 },
	{ 0, 0, 32 },

	{ 640, 480, 0 },
	{ 640, 480, 1 },
	{ 640, 480, 8 },
	{ 640, 480, 16 },
	{ 640, 480, 24 },
	{ 640, 480, 32 },

	{ 640, 400, 0 },
	{ 640, 400, 1 },
	{ 640, 400, 8 },
	{ 640, 400, 16 },
	{ 640, 400, 24 },
	{ 640, 400, 32 },

	{ 800, 600, 0 },
	{ 800, 600, 8 },
	{ 800, 600, 16 },
	{ 800, 600, 24 },
	{ 800, 600, 32 },

	{ 1024, 768, 0 },
	{ 1024, 768, 8 },
	{ 1024, 768, 16 },
	{ 1024, 768, 24 },
	{ 1024, 768, 32 },
	
	/* these are a last resort */

	{ 320, 200, 0 },
	{ 320, 200, 8 }
};


/* Read a single color register and split it into
** colors/transparent.  Return != 0 for invalid regno.
*/
static int kgifb_getcolreg(u_int regno, u_int *red, u_int *green, 
			   u_int *blue, u_int *transp, struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;

	if (regno > 255)
	{
		return 1;
	}

	*red   = kinfo->palette[regno].r;
	*green = kinfo->palette[regno].g;
	*blue  = kinfo->palette[regno].b;

	return 0;
}

/* Set a single color register.
** Return != 0 for invalid regno.
*/
static int kgifb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
			   u_int transp, struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;

	if (regno > 255) 
	{
		return 1;
	}

	kinfo->palette[regno].r = red;
	kinfo->palette[regno].g = green;
	kinfo->palette[regno].b = blue;
	
	if (regno < 16)
	{
#ifdef FBCON_HAS_CFB16
		kinfo->cmap_cfb15[regno] =
			((red & 0xf800) >> 1) | ((green & 0xf800) >> 6) | 
			((blue & 0xf800) >> 11);

		kinfo->cmap_cfb16[regno] =
			(red & 0xf800) | ((green & 0xfc00) >> 5) | 
			((blue & 0xf800) >> 11);
#endif
#ifdef FBCON_HAS_CFB24
		kinfo->cmap_cfb24[regno] =
			((red & 0xff00) << 8) | (green & 0xff00) |
			((blue & 0xff00) >> 8);
#endif
#ifdef FBCON_HAS_CFB32
		kinfo->cmap_cfb32[regno] = 
			((transp & 0xff00) << 16) | ((red & 0xff00) << 8) | 
			(green & 0xff00) | ((blue & 0xff00) >> 8);
#endif
	}

	return 0;
}

static int kgifb_setpalette(struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;

	kgi_clut clut;

	clut.size = 256;
	clut.data = kinfo->palette;

	if (kgifb_display.set_clut)
	{
		(kgifb_display.set_clut)(&kgifb_display, &clut, 
					 0, clut.size);
	}
	
	return 0;
}

/* Set the colormap
*/
static void do_install_cmap(int con, struct fb_info *info)
{
	int cmap_len = fb_display[con].var.bits_per_pixel;

	if (cmap_len <= 8) {
		cmap_len = 1 << cmap_len;
	} else {
		cmap_len = 16;
	}

	if (con != currcon)
	{
		return;
	}
	
	if (fb_display[con].cmap.len)
	{
		fb_set_cmap(&fb_display[con].cmap, 1, kgifb_setcolreg, info);
	}
	else
	{
		fb_set_cmap(fb_default_cmap(cmap_len), 1, 
			    kgifb_setcolreg, info);
	}
}

/* Open/release the frame buffer device
*/
static int kgifb_open(struct fb_info *info, int user)
{
	/* FIXME, but how /Marcus */
	kgifb_graph.dev.state |= KGI_MAPPED_ON_DISPLAY;

	DEBUG2("kgifb: open");
	MOD_INC_USE_COUNT;
	return(0);
}

static int kgifb_release(struct fb_info *info, int user)
{
	DEBUG2("kgifb: release");
	MOD_DEC_USE_COUNT;
	return(0);
}

static int kgifb_con_updatevar(int con, struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;
	struct fb_var_screeninfo *var;

	if (con == -1)
	{
		var = &kinfo->default_var;
	}
	else if (con == currcon)
	{
		var = &fb_display[con].var;
	}
	else
	{
		return 0;
	}

	/* update panning (needed for SCROLL_YPAN to work) */

	if (kgifb_display.set_origin)
	{
		(kgifb_display.set_origin)(&kgifb_display, 
			var->xoffset, var->yoffset);
	}

	return 0;
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
	switch (var->bits_per_pixel)
	{
	case 1:
	case 4:
	case 8:
	case 13:  /* <-- Hack city (for text modes) */

		/* These lengths are 16 because that's the resolution
		 * of kgi_color (passed to the set_clut function).
		 */
		var->red.offset = 0;
		var->red.length = 16;
		var->green.offset = 0;
		var->green.length = 16;
		var->blue.offset = 0;
		var->blue.length = 16;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;		

	case 15:
		var->red.offset = 10;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 5;
		var->blue.offset = 0;
		var->blue.length = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 16:
		var->red.offset = 11;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 6;
		var->blue.offset = 0;
		var->blue.length = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 24:
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 32:
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 24;
		var->transp.length = 8;
		break;
	}
	
	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;
}

static int convert_to_ggimode(kgi_modereq *dest, struct fb_var_screeninfo *var)
{
	int err=0;

	dest->frames = 1;
	dest->size.x = var->width;
	dest->size.y = var->height;
	dest->dpp.x = 1;
	dest->dpp.y = 1;
	
	switch (var->bits_per_pixel) {
		case 1:  dest->graphtype = KGIGT_1BIT;  break;
		case 4:  dest->graphtype = KGIGT_4BIT;  break;
		case 8:  dest->graphtype = KGIGT_8BIT;  break;
		case 15: dest->graphtype = KGIGT_15BIT; break;
		case 16: dest->graphtype = KGIGT_16BIT; break;
		case 24: dest->graphtype = KGIGT_24BIT; break;
		case 32: dest->graphtype = KGIGT_32BIT; break;

		/* Oh this is such a hack: */
		case 13: dest->graphtype = KGIGT_TEXT16;
			 dest->dpp.x = 8; dest->dpp.y = 16; 
			 break;

		default:
			DEBUG1("Unknown bpp '%d'.", var->bits_per_pixel);
			
			/* Just resort to 8 bit (better than nothing) */
			dest->graphtype = KGIGT_8BIT;

			err = -1;
	}
	
	dest->visible.x = var->xres / dest->dpp.x;
	dest->visible.y = var->yres / dest->dpp.y;
	dest->virt.x = var->xres_virtual / dest->dpp.x;
	dest->virt.y = var->yres_virtual / dest->dpp.y;
	
	return err;
}

static int convert_from_ggimode(kgi_modereq *src, struct fb_var_screeninfo *var)
{
	int err=0;

	var->xres = src->visible.x * src->dpp.x;
	var->yres = src->visible.y * src->dpp.y;
	var->xres_virtual = src->virt.x * src->dpp.x;
	var->yres_virtual = src->virt.y * src->dpp.y;
	var->width  = src->size.x;
	var->height = src->size.y;

	var->grayscale = 0;
	var->nonstd = 0;
	var->accel_flags = 0;

	switch (src->graphtype) {
		case KGIGT_1BIT:  var->bits_per_pixel = 1;  break;
		case KGIGT_4BIT:  var->bits_per_pixel = 4;  break;
		case KGIGT_8BIT:  var->bits_per_pixel = 8;  break;
		case KGIGT_15BIT: var->bits_per_pixel = 15; break;
		case KGIGT_16BIT: var->bits_per_pixel = 16; break;
		case KGIGT_24BIT: var->bits_per_pixel = 24; break;
		case KGIGT_32BIT: var->bits_per_pixel = 32; break;

		/* Oh this is such a hack: */
		case KGIGT_TEXT16: var->bits_per_pixel = 13; break;

		default:
			DEBUG1("Unknown graphtype '%08x'.", src->graphtype);
			
			/* Just resort to 8 bit (better than nothing) */
			var->bits_per_pixel = 8;

			err = -1;
	}
	
	set_color_bitfields(var);

	return err;
}

static int convert_from_ggitimings(struct kgi_mode *mode, 
				   struct fb_var_screeninfo *var)
{
	var->pixclock = 1000000000 / (mode->dclk / 1000);
	
	var->hsync_len = mode->x.syncend - mode->x.syncstart;
	var->vsync_len = mode->y.syncend - mode->y.syncstart;
	var->left_margin = mode->x.total - mode->x.syncend;
	var->right_margin = mode->x.syncstart - mode->x.width;
	var->upper_margin = mode->y.total - mode->y.syncend;
	var->lower_margin = mode->y.syncstart - mode->y.width;
	
	return 0;
}

extern int console_loglevel;

static void kgifb_encode_fix(struct fb_fix_screeninfo *fix, 
			     struct fb_var_screeninfo *var)
{
	memset(fix, 0, sizeof(struct fb_fix_screeninfo));

	strcpy(fix->id, "KGI");

#ifdef LINUX_23	
	fix->smem_start  = the_info.videophys;
	fix->mmio_start  = the_info.mmio_start;
#else
	fix->smem_start  = (void*)the_info.videophys;
	fix->mmio_start  = (void*)the_info.mmio_start;
#endif
	fix->smem_len    = the_info.videosize;
	fix->mmio_len    = the_info.mmio_len;
	fix->accel       = the_info.accel;

	fix->visual      = FB_VISUAL_PSEUDOCOLOR;
	fix->type        = FB_TYPE_PACKED_PIXELS;
	fix->type_aux    = 0;
	
	fix->xpanstep    = 1;
	fix->ypanstep    = 1;
	fix->ywrapstep   = 0;
	fix->line_length = get_stride(var->xres_virtual, 
					var->bits_per_pixel);

	switch (var->bits_per_pixel)
	{
	case 15: case 16:
	case 24: case 32:

		fix->visual = FB_VISUAL_DIRECTCOLOR;
		break;

	case 13:  /* <-- welcome to Hack City (tm) */

		fix->visual   = FB_VISUAL_PSEUDOCOLOR;
		fix->type     = FB_TYPE_TEXT;
		fix->type_aux = FB_AUX_TEXT_CGA;
		fix->xpanstep = 8;
		fix->ypanstep = 16;
		break;
	}
}

static int kgifb_get_fix(struct fb_fix_screeninfo *fix, int con,
			 struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;
	struct fb_var_screeninfo *var;
	
	if (con == -1)
	{
		var = &kinfo->default_var;
	}
	else
	{
		var = &fb_display[con].var;
	}
	
	kgifb_encode_fix(fix, var);
	
	return 0;
}

static int kgifb_get_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;

	if (con == -1)
	{
		*var = kinfo->default_var;
	}
	else
	{
		*var = fb_display[con].var;
	}
	
	return 0;
}


static void get_framebuffer_info(void)
{
	struct kgi_mmio_region *temp = kgifb_display.mode.mmio;

	the_info.videovirt = (unsigned long) kgifb_display.mode.fb;
	
	while (temp != NULL) {
		if (temp->type == MMIO_FRAMEBUFFER) {
			the_info.videophys = temp->pstart;
			the_info.videosize = temp->psize;
			break;
		}
		temp = temp->next;
	}
	if (temp == NULL) {
		DEBUG1("get_framebuffer_info: mmio == NULL");
		/* The next is just best effort */
		the_info.videophys = virt_to_phys((void *)the_info.videovirt);
		the_info.videosize = kgifb_display.mode.fb_size;
	}

	/* For KGI mmio accel */
	if (kgifb_display.mode.fbcon_mmio_start != 0) {
		the_info.mmio_start = kgifb_display.mode.fbcon_mmio_start;
		the_info.mmio_len = kgifb_display.mode.fbcon_mmio_len;
	} else {
		the_info.mmio_start = 0;
		the_info.mmio_len = 0;
	}
	the_info.accel = kgifb_display.mode.fbcon_accel;

#ifdef CONFIG_MTRR
	if (kgicon_mtrr) {
		/* Some stupid drivers don't report the entire frame buffer,
		   so we have to re-create the MTRR entry every time.
		*/
		if (the_info.mtrr_entry >= 0) {
			mtrr_del(the_info.mtrr_entry, 0, 0);
		}
		the_info.mtrr_entry
			= mtrr_add(the_info.videophys, the_info.videosize,
				   MTRR_TYPE_WRCOMB, 1);
	}
#endif
}

static int kgi_do_check_mode(struct fb_var_screeninfo *var) {

	struct kgi_mode newmode;

	convert_to_ggimode(&newmode.request, var);

	if ((kgifb_display.check_mode)(&kgifb_display,&newmode,CMD_PROPOSE)) {

		DEBUG1("Checking failed.");
		printk("Checking failed.");
		return -EINVAL;
	}

	DEBUG1("Checking success: %dx%d#%dx%d D%dx%d [0x%x]",
		newmode.request.visible.x, newmode.request.visible.y,
		newmode.request.virt.x, newmode.request.virt.y,
		newmode.request.dpp.x, newmode.request.dpp.y,
		newmode.request.graphtype);

	convert_from_ggimode(&newmode.request, var);
	convert_from_ggitimings(&newmode, var);

	return 0;
}

static int kgi_do_set_mode(struct fb_var_screeninfo *var)
{
	struct kgi_mode newmode;

	convert_to_ggimode(&newmode.request, var);
	
	if ((kgifb_display.check_mode)(&kgifb_display,&newmode,CMD_PROPOSE)) 
	{
		DEBUG1("Activating failed.");

		return -EINVAL;
	}

	(kgifb_display.set_mode)(&kgifb_display, &newmode);

	if (kgifb_display.set_split)
	{
		/* NOTE: slight hack here to multiply by magnify.y */

		(kgifb_display.set_split)(&kgifb_display, 
			newmode.request.visible.y *
			newmode.request.dpp.y * newmode.magnify.y);
	}

	DEBUG1("Activating success: %dx%d#%dx%d D%dx%d [0x%x]",
		newmode.request.visible.x, newmode.request.visible.y,
		newmode.request.virt.x, newmode.request.virt.y,
		newmode.request.dpp.x, newmode.request.dpp.y,
		newmode.request.graphtype);

	convert_from_ggimode(&newmode.request, var);
	convert_from_ggitimings(&newmode, var);

	get_framebuffer_info();
	
	memcpy(&kgifb_graph.dev.mode, &newmode.request,
	       sizeof(newmode.request));

	return 0;
}

static void kgi_do_set_font(struct fb_var_screeninfo *var)
{
	if ((var->bits_per_pixel == 13) &&  /* <-- text mode hack */
	    kgifb_display.set_font && kgicon_font.data) {
		
		(kgifb_display.set_font)(&kgifb_display,
			&kgicon_font, 0, 255);
	}
}	

static int kgifb_set_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info)
{
	struct fb_info_kgicon *kinfo = (struct fb_info_kgicon *) info;

	int err, activate = var->activate;
	
	struct display *display;
	struct display_switch *new_sw;
	void *new_data = NULL;
	
	struct fb_var_screeninfo old;
	struct fb_fix_screeninfo fix;
	

	if (con == -1)
	{
		/* used during initialization */
		display = &kinfo->default_display;	
	}
	else
	{
		display = &fb_display[con];
	}
	
	
	/* FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
	** as FB_VMODE_SMOOTH_XPAN is only used internally
	*/

	if (var->vmode & FB_VMODE_CONUPDATE) 
	{
		var->vmode |= FB_VMODE_YWRAP;
		var->xoffset = display->var.xoffset;
		var->yoffset = display->var.yoffset;
	}

	/* Some very basic checks
	 */
	if (!var->xres || !var->yres)
	{
		return -EINVAL;
	}
	
	if (var->xres > var->xres_virtual)
	{
		var->xres_virtual = var->xres;
	}
	
	if (var->yres > var->yres_virtual)
	{
		var->yres_virtual = var->yres;
	}
	
	if (var->bits_per_pixel <= 1)
	{
		var->bits_per_pixel = 1;
	}

	else if (var->bits_per_pixel <= 4)
	{
		var->bits_per_pixel = 4;
	}
	
	else if (var->bits_per_pixel <= 8)
	{
		var->bits_per_pixel = 8;
	}
	
	else if (var->bits_per_pixel == 13)
	{
		/* This is a hack for text mode */
	}

	else if (var->bits_per_pixel <= 15)
	{
		var->bits_per_pixel = 15;
	}
	
	else if (var->bits_per_pixel <= 16)
	{
		var->bits_per_pixel = 16;
	}
	
	else if (var->bits_per_pixel <= 24)
	{
		var->bits_per_pixel = 24;
	}
	
	else if (var->bits_per_pixel <= 32)
	{
		var->bits_per_pixel = 32;
	}
	
	else
	{
		return -EINVAL;
	}
	
	set_color_bitfields(var);

	switch(activate & FB_ACTIVATE_MASK) {
		case FB_ACTIVATE_NOW:
			break;
		case FB_ACTIVATE_TEST:
			printk("Testing kgicon mode %dx%d#%dx%d[%d].\n",
				var->xres, var->yres, 
				var->xres_virtual, var->yres_virtual, 
				var->bits_per_pixel);
			return kgi_do_check_mode(var);
		default:
			DEBUG1("Strange activation mode %x: kgicon mode %dx%d#%dx%d[%d].\n",
				activate,
				var->xres, var->yres, 
				var->xres_virtual, var->yres_virtual, 
				var->bits_per_pixel);
			return 0;
	}

	DEBUG1("Activating kgicon mode %dx%d#%dx%d[%d].",
		var->xres, var->yres, 
		var->xres_virtual, var->yres_virtual, 
		var->bits_per_pixel);

	old = display->var;

	if (old.xres == var->xres && 
	    old.yres == var->yres &&
	    old.xres_virtual == var->xres_virtual && 
	    old.yres_virtual == var->yres_virtual &&
	    old.bits_per_pixel == var->bits_per_pixel) 
	{
		/* no change */
	
		kgifb_con_updatevar(con, info);
		kgifb_setpalette(info);
		return 0;
	}

	/* work out new console renderer */

	switch (var->bits_per_pixel) 
	{
#undef KGICON_HAVE_A_BPP
#ifdef FBCON_HAS_MFB
#define KGICON_HAVE_A_BPP 1
	case 1:
		new_sw = &fbcon_mfb;
		break;
#endif
#ifdef FBCON_HAS_CFB2
#define KGICON_HAVE_A_BPP 1
	case 2:
		new_sw = &fbcon_cfb2;
		break;
#endif
#ifdef FBCON_HAS_CFB4
#define KGICON_HAVE_A_BPP 1
	case 4:
		new_sw = &fbcon_cfb4;
		break;
#endif
#ifdef FBCON_HAS_CFB8
#define KGICON_HAVE_A_BPP 1
	case 8:
		new_sw = &fbcon_cfb8;
		break;
#endif
#ifdef FBCON_HAS_CFB16
#define KGICON_HAVE_A_BPP 1
	case 15:
		new_sw = &fbcon_cfb16;
		new_data = kinfo->cmap_cfb15;
		break;

	case 16:
		new_sw = &fbcon_cfb16;
		new_data = kinfo->cmap_cfb16;
		break;
#endif
#ifdef FBCON_HAS_CFB24
#define KGICON_HAVE_A_BPP 1
	case 24:
		new_sw = &fbcon_cfb24;
		new_data = kinfo->cmap_cfb24;
		break;
#endif
#ifdef FBCON_HAS_CFB32
#define KGICON_HAVE_A_BPP 1
	case 32:
		new_sw = &fbcon_cfb32;
		new_data = kinfo->cmap_cfb32;
		break;
#endif
#ifdef FBCON_HAS_VGA
	case 13:  /* welcome to Hack City (tm) */
		new_sw = &fbcon_vga;
		break;
#endif
#ifndef KGICON_HAVE_A_BPP
#error You do not have proper fbcon support in your kernel! Please rebuild your kernel and make sure you set CONFIG_FBCON_ADVANCED to NO.
#endif
	default:
		ERROR("kgifb_set_var: Unsupported bpp '%d'.", 
			var->bits_per_pixel);
		return -1;
	}
	
	/* set mode */

	if ((con == -1) || (con == currcon)) {
		if (kgi_do_set_mode(var) < 0) {
			return -EINVAL;
		}
	}
	
	/* update display structure */

	kgifb_encode_fix(&fix, var);

	display->screen_base = (char *) kinfo->videovirt;
	display->visual = fix.visual;
	display->type = fix.type;
	display->type_aux = fix.type_aux;
	display->ypanstep = fix.ypanstep;
	display->ywrapstep = fix.ywrapstep;
	display->line_length = fix.line_length;
	display->next_line = fix.line_length;
	display->can_soft_blank = 0;   /* Not supported yet */
	display->inverse = kgicon_inverse;
	
	display->var = *var;
	display->dispsw = new_sw;
	display->dispsw_data = new_data;

	if (var->yres == var->yres_virtual) {
		display->scrollmode = SCROLL_YREDRAW;
	} else {
		display->scrollmode = 0;  /* SCROLL_YPAN */
	}

	if (old.bits_per_pixel != var->bits_per_pixel) 
	{
		/* NOTE: this _frees_ the colormap */
		
		if ((err = fb_alloc_cmap(&display->cmap, 0, 0)))
		{
			return err;
		}
		
		if (con != -1) {
			do_install_cmap(con, info);
		}
	}

	kgifb_con_updatevar(con, info);
	kgifb_setpalette(info);

	if ((con != -1) && info->changevar)
	{
		DEBUG2("kgicon: Changing var...");

		(*info->changevar)(con);
	}

	/* set font */

	kgicon_font.data = display->fontdata;
	
	if ((con == -1) || (con == currcon)) {
		kgi_do_set_font(var);
	}

	return 0;
}
	
static int kgifb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			  struct fb_info *info)
{
	int cmap_len = fb_display[con].var.bits_per_pixel;

	if (cmap_len <= 8) {
		cmap_len = 1 << cmap_len;
	} else {
		cmap_len = 16;
	}


	if (con == currcon)
	{
		return fb_get_cmap(cmap, kspc, kgifb_getcolreg, info);
	}
	else if (fb_display[con].cmap.len) /* non default colormap? */
	{
		
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	}
	else
	{
		fb_copy_cmap(fb_default_cmap(cmap_len),
			     cmap, kspc ? 0 : 2);
	}
	
	return 0;
}

static int kgifb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			  struct fb_info *info)
{
	int err=0;
	int cmap_len = fb_display[con].var.bits_per_pixel;

	if (cmap_len <= 8) {
		cmap_len = 1 << cmap_len;
	} else {
		cmap_len = 16;
	}

	if (! fb_display[con].cmap.len) 
	{
		/* No colormap allocated? */
		
		err = fb_alloc_cmap(&fb_display[con].cmap, cmap_len, 0);

		if (err) return err;
	}

	if (con == currcon) 
	{
		err = fb_set_cmap(cmap, kspc, kgifb_setcolreg, info);

		kgifb_setpalette(info);
	} 
	else 
	{
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	}

	return 0;
}

static int kgifb_pan_display(struct fb_var_screeninfo *var, int con,
			     struct fb_info *info)
{
	if (con < 0)
		return 0;

	if ((var->vmode & FB_VMODE_YWRAP) &&
	    ((var->yoffset < 0) || 
	     (var->yoffset >= fb_display[con].var.yres_virtual) || 
	     (var->xoffset < 0) ||
	     (var->xoffset >= fb_display[con].var.xres_virtual)))
	{
		return -EINVAL;
	}

	fb_display[con].var.xoffset = var->xoffset;
	fb_display[con].var.yoffset = var->yoffset;
	
	fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;
	fb_display[con].var.vmode |= var->vmode & FB_VMODE_YWRAP;

	kgifb_con_updatevar(con, info);
	
	return 0;
}


static int driver_ioctl(unsigned int cmd, unsigned long arg)
{
	struct kgi_mode newmode;
	int size = 0, ret = 0;
	void *ptr = NULL;

	switch (cmd) {
	case DRIVER_TESTDRIVER:
		return 0;

	case DRIVER_GETINFO:
		ptr = &fbcon_driver_id;
		size = sizeof(fbcon_driver_id);
		goto copy_and_ret;

	case DRIVER_SETINFO:
		/* It is usually not permissible to upload the driver
		   information. */
		return -EINVAL;

	case DRIVER_TESTMODE:
		memcpy(&newmode.request, kgifb_graph.io_buf,
		       sizeof(newmode.request));
		if ((kgifb_display.check_mode)(&kgifb_display, &newmode,
					       CMD_PROPOSE)) {
			DEBUG1("Checking failed.");
			ptr = kgifb_graph.io_res;
			size = kgifb_graph.io_cnt;
			ret = -EINVAL;
		}
		goto copy_and_ret;

	case DRIVER_SETGRAPHMODE:
		memcpy(&newmode.request, kgifb_graph.io_buf,
		       sizeof(newmode.request));
		if ((kgifb_display.check_mode)(&kgifb_display, &newmode,
					       CMD_PROPOSE)) {
			ptr = kgifb_graph.io_res;
			size = kgifb_graph.io_cnt;
			ret = -EINVAL;
			goto copy_and_ret;
		}
		(kgifb_display.set_mode)(&kgifb_display, &newmode);

		if (kgifb_display.set_split) {
			(kgifb_display.set_split)(&kgifb_display, 
						  newmode.request.visible.y *
						  newmode.request.dpp.y *
						  newmode.magnify.y);
		}
		ptr = kgifb_graph.io_res;
		size = kgifb_graph.io_cnt;
		goto copy_and_ret;

	case DRIVER_GETGRAPHMODE:
		ptr = &kgifb_graph.dev.mode;
		size = sizeof(kgifb_graph.dev.mode);
		goto copy_and_ret;
	}

	NOTICE("kgifb_ioctl: unknown IOC_DRIVER type ioctl().");
	return -EINVAL;

  copy_and_ret:
	if (copy_to_user((void *)arg, ptr, size)) {
		ret = -EFAULT;
	}

	return ret;
}


static int kgifb_ioctl(struct inode *inode, struct file *file, 
		       unsigned int cmd, unsigned long arg, int con,
		       struct fb_info *info)
{
	int ret = 0;
	int type;

	if (cmd == KGICON_PROBE) {
		if (put_user(KGICON_PROBE_MAGIC, (int*)arg)) return -EFAULT;
		return 0;
	}

	if ((kgifb_graph.io_cnt = _IOC_SIZE(cmd)) > KGICON_IO_BUF_SIZE) {
		return -EINVAL;
	}
	kgifb_graph.io_res = NULL;

	if (cmd & IOC_IN) {
		if (copy_from_user(kgifb_graph.io_buf,(void *) arg,
				   kgifb_graph.io_cnt))
			return -EFAULT;
	}

	type = _IOC_TYPE(cmd);

	if (type >= IOC_DRIVER && type <= IOC_DRIVER+KGI_NR_SUBSYSTEMS) {
		switch (type) {
		case IOC_DRIVER:
			return driver_ioctl(cmd, arg);
		default:
			type -= KGI_IOCTLBASE;
			ret = ((type < KGI_NR_SUBSYSTEMS)
			       && (kgifb_display.cmd[type])
			       && kgifb_display.cmd[type](&kgifb_graph,
							  cmd, arg) == 0)
				? 0 : -EINVAL;
			if ((cmd & IOC_OUT) && ret == 0) {
				if (kgifb_graph.io_res &&
				    0 < kgifb_graph.io_cnt) {
					ret = copy_to_user((void *)arg,
							   kgifb_graph.io_res,
							   kgifb_graph.io_cnt)
						? -EFAULT : 0;
				} else {
					ERROR("Got here!");
					ret = -EINVAL;
				}
			}

			return ret;
		}
	}

	NOTICE("kgifb_ioctl: 0x%08x not supported", cmd);
	return -EINVAL;
}


#ifdef KGICON_USE_MMAP

#ifdef USE_VMOPS

#ifndef MAP_NR
# define MAP_NR(addr)	(__pa(addr) >> PAGE_SHIFT)
#endif
#ifndef get_page
# define get_page(p)	atomic_inc(&(p)->count)
#endif

#ifdef LINUX_23
static struct page *
#else
static unsigned long
#endif
graph_GC_nopage(struct vm_area_struct *vma, 
		unsigned long addr, int no_share)
{
	/* Ensure that the page usage count is incremented! */
	get_page(&mem_map[MAP_NR(kgifb_graph.gc)]);

#ifdef LINUX_23
	return &mem_map[MAP_NR(kgifb_graph.gc)];
#else
	return (unsigned long) kgifb_graph.gc;
#endif
}

static struct vm_operations_struct graph_GC_vmops = {
	NULL,                   /* open		*/
	NULL,			/* close	*/
	NULL,			/* unmap	*/
	NULL,                   /* protect	*/
	NULL,                   /* sync		*/
#ifdef HAVE_STRUCT_VM_OPERATIONS_ADVISE
	NULL,                   /* advise	*/
#endif
	graph_GC_nopage,	/* nopage	*/
	NULL,                   /* wppage	*/
	NULL,                   /* swapout	*/
#ifndef LINUX_23
	NULL,                   /* swapin	*/
#endif
};
#endif


#define PROC_CONSOLE() (-1)

static int kgifb_mmap(struct fb_info *info, struct file *file, struct vm_area_struct *vma)
{
	size_t size = vma->vm_end - vma->vm_start;
	unsigned long off;

#ifdef HAVE_VMA_VM_PGOFF
	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT)) return -EINVAL;
	off = vma->vm_pgoff << PAGE_SHIFT;
#else
	off = vma->vm_offset;
#endif

	if ((off & MMAP_TYPE) == MMAP_TYPE_GC) {
		if (size != PAGE_SIZE) {
			return -EINVAL;
		}

		kgifb_graph.gc->version = 0;
		kgifb_graph.gc->oldversion = 0;

#ifdef USE_VMOPS
		vma->vm_ops = &graph_GC_vmops;
#else
		if (remap_page_range(vma->vm_start,
				     virt_to_phys(kgifb_graph.gc),
				     size, vma->vm_page_prot)) {
			return -EAGAIN;
		}
#endif
		vma->vm_file = file;
#ifdef LINUX_23
		atomic_inc(&file->f_count);
#else
		file->f_count++;
#endif

		return EOK;
	} else {
		struct fb_fix_screeninfo fix;
		struct fb_var_screeninfo var;
		kgi_addr start;
		kgiu32 length;

		info->fbops->fb_get_fix(&fix, PROC_CONSOLE(), info);
		start = (kgi_addr) fix.smem_start;
		length = (start & ~PAGE_MASK) + fix.smem_len;
		start &= PAGE_MASK;
		length = (length + ~PAGE_MASK) & PAGE_MASK;

		if (off >= length) {
			/* MMIO region */
#ifndef NONROOT_MMIO
			/* Only allow super user to map MMIO region */
			if (!suser()) return -EPERM;
#endif /* NONROOT_MMIO */
			info->fbops->fb_get_var(&var, PROC_CONSOLE(), info);
			if (var.accel_flags) return -EINVAL;
			off -= length;
			start = (kgi_addr) fix.mmio_start;
			length = (start & ~PAGE_MASK) + fix.mmio_len;
			start &= PAGE_MASK;
			length = (length + ~PAGE_MASK) & PAGE_MASK;
		}

		if ((vma->vm_end - vma->vm_start + off) > length) {
			return -EINVAL;
		}

		off += start;
#ifdef HAVE_VMA_VM_PGOFF
		vma->vm_pgoff = off >> PAGE_SHIFT;
#else
		vma->vm_offset = off;
#endif

#if defined(__mc68000__)
		if (CPU_IS_020_OR_030) {
			pgprot_val(vma->vm_page_prot) |= _PAGE_NOCACHE030;
		}
		if (CPU_IS_040_OR_060) {
			pgprot_val(vma->vm_page_prot) &= _CACHEMASK040;
			pgprot_val(vma->vm_page_prot) |= _PAGE_NOCACHE_S;
		}
#elif defined(__i386__)
		if (boot_cpu_data.x86 > 3) {
			pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
		}
#elif defined(__mips__)
		pgprot_val(vma->vm_page_prot) &= ~_CACHE_MASK;
		pgprot_val(vma->vm_page_prot) |= _CACHE_UNCACHED;
#elif defined(__powerpc__)
		pgprot_val(vma->vm_page_prot) |= _PAGE_GUARDED|_PAGE_NO_CACHE;
#elif defined(__alpha__)
		/* No caching is done in the I/O space quadrant. */
#elif defined(__sparc__)
#error What do Sparcs need here???
#else
#error Compiling for unknown architecture, please check this code!
#endif
		if (remap_page_range(vma->vm_start, off,
				     vma->vm_end - vma->vm_start,
				     vma->vm_page_prot)) {
			return -EAGAIN;
		}
#ifdef LINUX_23
		atomic_inc(&file->f_count);
#else
		file->f_count++;
#endif
		vma->vm_file = file;

		return 0;
	}
	return -EINVAL;
}

#endif /* KGICON_USE_MMAP */


static struct fb_ops kgifb_ops = 
{
#ifdef LINUX_24
	NULL,
#endif
	kgifb_open,
	kgifb_release,
	kgifb_get_fix,
	kgifb_get_var,
	kgifb_set_var,
	kgifb_get_cmap,
	kgifb_set_cmap,
	kgifb_pan_display,
	kgifb_ioctl,
#ifdef KGICON_USE_MMAP
	kgifb_mmap		/* fb_mmap */
#else
	NULL
#endif
};

void kgifb_setup(char *options, int *ints)
{
	char *this_opt;

	if (!options || !*options) {
		return;
	}
     
	for(this_opt = strtok(options,","); this_opt; 
	    this_opt = strtok(NULL,",")) {
		if (!*this_opt) {
			continue;
		}

		if (!strcmp(this_opt, "inverse")) {
			kgicon_inverse = 1;
		}
#ifdef ENABLE_FONT_SELECTION_VIA_BOOT_OPTION
		else if (!strncmp(this_opt, "font:", 5)) {
			strcpy(fb_info.fontname, this_opt + 5);
			NOTICE("kgifb_setup: option %s", this_opt);
		}
#endif
	}
}

static int kgifb_con_switch(int con, struct fb_info *info)
{
	/* Do we have to save the colormap? */

	if (fb_display[currcon].cmap.len) {
		fb_get_cmap(&fb_display[currcon].cmap, 1, 
			    kgifb_getcolreg, info);
	}

	currcon = con;

	kgi_do_set_mode(&fb_display[con].var);
	kgi_do_set_font(&fb_display[con].var);
	kgifb_con_updatevar(con, info);

	do_install_cmap(con, info);
	kgifb_setpalette(info);

	return 0;
}

/* 0 unblank, 1 blank, 2 no vsync, 3 no hsync, 4 off */
static void kgifb_con_blank(int blank, struct fb_info *info)
{
	/* Not supported yet */
}

/* Calculate the stride (bytes per line) of the display */
static u_long get_stride(int xres_virtual, int bpp)
{
	if (bpp == 13) {
		/* <-- text mode hack */
		return (xres_virtual / 8) * 2;
	}
	
	if (bpp < 8) {
		return (xres_virtual * bpp + 7) / 8;
	} else {
		return xres_virtual * ((bpp+7) / 8);
	}
}


int __init kgifb_init(void)
{
	int err;
	int i;

	DEBUG2("kgifb_init: Start");

	memset(&the_info, 0, sizeof(struct fb_info_kgicon));

	err = init_kgi();
	if (err != 0) {
		ERROR("init_kgi() failed, bailing out....");
		return err;
	}

	/* Let there be consoles.... */
	strcpy(the_info.info.modename, "KGI");

	the_info.info.node        = -1;
	the_info.info.fbops       = &kgifb_ops;
	the_info.info.disp        = &the_info.default_display;
	the_info.info.switch_con  = &kgifb_con_switch;
	the_info.info.updatevar   = &kgifb_con_updatevar;
	the_info.info.blank       = &kgifb_con_blank;
	the_info.info.flags       = FBINFO_FLAG_DEFAULT;
	the_info.info.changevar   = NULL;
	the_info.info.display_fg  = NULL;

	/* Text mode hack.  Fbcon currently doesn't handle text modes
	 * very well, and we need a way to force fbcon's notion of font
	 * size to be something fixed, such as 8x16
	 */
	strcpy(the_info.info.fontname, "VGA8x16");

	/* Set the initial mode...
	 */
	 
	for (err = -1, i = 0; err && i < NR_DEFAULT_MODES; i++) {
		int w   = default_modes[i].width  ? default_modes[i].width
				: kgicon_width;
		int h   = default_modes[i].height ? default_modes[i].height
				: kgicon_height;
		int bpp = default_modes[i].bpp    ? default_modes[i].bpp 
				: kgicon_bpp;

		the_info.default_var.xres           = w;
		the_info.default_var.yres           = h;
		the_info.default_var.xres_virtual   = w;
		the_info.default_var.yres_virtual   = h;
		the_info.default_var.bits_per_pixel = bpp;
		the_info.default_var.activate       = FB_ACTIVATE_NOW;
	
		err = kgifb_set_var(&the_info.default_var, -1,
				    &the_info.info);
	}

	if (err) {
		ERROR("kgicon: Cannot find any usable video mode.");
		exit_kgi();
		return err;
	}
	
	NOTICE("kgifb_init: xres %d yres %d bpp %d vaddr 0x%lx paddr 0x%lx"
		" size %ld", the_info.default_var.xres, 
		the_info.default_var.yres, 
		the_info.default_var.bits_per_pixel, 
		the_info.videovirt, the_info.videophys, the_info.videosize);


	err = register_framebuffer(&the_info.info);
	if (err < 0) {
		ERROR("Unable to register framebuffer device.");
		exit_kgi();
		return err;
	}

	NOTICE("fb%d: %s frame buffer device using %ldK of video memory",
	       GET_FB_IDX(the_info.info.node), the_info.info.modename, 
	       the_info.videosize >> 10);
	
	err = kgiproc_init();
	if (err) {
		printk("Unable to init /proc interface.\n");
		unregister_framebuffer(&the_info.info);
		exit_kgi();
		return err;
	}

	kgisysctl_init(GET_FB_IDX(the_info.info.node));

	return 0;
}

static void kgifb_exit(void)
{
	int fb_num = GET_FB_IDX(the_info.info.node);

	kgiproc_exit();

	kgisysctl_exit();

	unregister_framebuffer(&the_info.info);
	
	NOTICE("fb%d: unregistered.", fb_num);
	
	exit_kgi();
}


#ifndef THIS_MODULE
/* Make sure that the module isn't removed while the
   /proc/sys/dev/ directorytree is being used. */
static void sysctl_modcount(struct inode *inode, int fill)
{
	if (fill) MOD_INC_USE_COUNT;
	else MOD_DEC_USE_COUNT;
}
#endif


static int sysctl_dosubsys(struct kgi_subparam *(*func)(void),
			    char *name, int idx)
{
	kgi_subparam *params;
		
	if ((params = func()) != NULL) {
		int num_params = 0;
		int i;
		struct ctl_table *c_leafs;

		while (params[num_params].handle != 0)
			num_params++;

		c_leafs = kmalloc((num_params+1)*sizeof(struct ctl_table),
				  GFP_KERNEL);
		if (! c_leafs) {
			return idx;
		}
		memset(c_leafs, 0, (num_params+1)*sizeof(struct ctl_table));
		for (i = 0; i < num_params; i++) {
			c_leafs[i].ctl_name = params[i].handle;
			c_leafs[i].procname = params[i].name;
			c_leafs[i].data     = params[i].param;
			c_leafs[i].maxlen   = params[i].size;
			c_leafs[i].mode     = params[i].mode;

			switch (params[i].type) {
			case KGI_SUBPARAM_INT:
				c_leafs[i].proc_handler = &proc_dointvec;
				break;
			case KGI_SUBPARAM_STRING:
				c_leafs[i].proc_handler = &proc_dostring;
				c_leafs[i].strategy = &sysctl_string;
				break;
			default:
				/* If we get here the subsystem made an error,
				   so we just terminate the list here */
				c_leafs[i].ctl_name = 0;
				break;
			}
		}
		
		memset(sysctl_subsys+idx, 0, sizeof(struct ctl_table));
		sysctl_subsys[idx].ctl_name = idx+1;
		sysctl_subsys[idx].procname = name;
		sysctl_subsys[idx].child    = c_leafs;
		sysctl_subsys[idx].mode     = 0555;

		idx++;
	}

	return idx;
}

static void kgisysctl_init(int fbidx)
{
	int idx = KGI_SUBSYS_LEAFS;
	
	/* If we have more than 999999999 framebuffer devices
	   we'd write past the end of the string... */
	if (fbidx > 999999999) {
		ERROR("Your system is evil. Kill some framebuffers. NOW.");
		return;
	}

	sysctl_fbnum->ctl_name = SYSCTL_KGICON_BASE+fbidx+1;
	sprintf(sysctl_fb_string, "fb%d", fbidx);
	
	
	idx = sysctl_dosubsys(kgim_monitor_get_params, "monitor", idx);
	/* 
	   ...
	   Add more subsystems here
	   ...
	*/
	sysctl_subsys[idx].ctl_name = 0;
	
	if (idx > 0) {
		sysctl_th = register_sysctl_table(sysctl_fbnum, 0);
#ifdef THIS_MODULE
		sysctl_fbnum->de->owner = THIS_MODULE;
#else
		sysctl_fbnum->de->fill_inode = &sysctl_modcount;
#endif
	}
}

static void kgisysctl_exit(void)
{
	if (sysctl_th != NULL) {
		int i;
		unregister_sysctl_table(sysctl_th);
		for (i = 0; sysctl_subsys[i].ctl_name != 0; i++) {
			kfree(sysctl_subsys[i].child);
		}
	}
}


#ifdef KGICON_USE_PROC
/*	Ping Pong Buffering
*/

static int pingpong = 0;


static unsigned long graph_accel_nopage(struct vm_area_struct *vma, unsigned long addr, int no_share)
{
/*	kgiu32 offset = addr - vma->vm_start; */
	kgiu32 size = accel_map.ping.size;
	kgiu32 domapstart, unmapstart;
	struct kgi_accel_buffer *domap, *unmap;
	
	ASSERT(accel_map.ping.size == accel_map.pong.size);

	DEBUG2("%s:%i: accel '%s' nopage @%.8lx\n",__FILE__,__LINE__,
	       accel_map.accel->name, addr);

	if (!(accel_map.file && (kgifb_graph.dev.state & KGI_MAPPED_ON_DISPLAY))) 
	{
		ERROR("graph_accel_nopage: fault on unmapped display");
		/* !!!	well, this should be done in a better way.
		*/
		return 1;
	}
	
	/*	figure out which buffer to unmap and which to map. Wait until
	**	it is processed before continuing.
	*/
/*	if (offset < size) */
	if (pingpong == 0)
	{
		if (accel_map.ping.state != ACCEL_IDLE) 
		{
			accel_map.accel->wait(accel_map.accel);
		}
		unmap = &(accel_map.pong);
		domap = &(accel_map.ping);
		unmapstart = (domapstart = vma->vm_start) + size;
	} else 
	{
		if (accel_map.pong.state != ACCEL_IDLE) 
		{		
			accel_map.accel->wait(accel_map.accel);
		}
		unmap = &(accel_map.ping);
		domap = &(accel_map.pong);
		domapstart = (unmapstart = vma->vm_start) + size;
	}
	
	domap->state = ACCEL_IDLE;
	
	DEBUG2("%s:%i: domap %p state %i, unmap %p state %i\n",
	       __FILE__,__LINE__, domap->buffer, domap->state, unmap->buffer,
	       unmap->state);
	
	if (unmap->state == ACCEL_FILL) 
	{
		DEBUG2("%s:%i: unmap %.8lx size %lx from %.8lx\n",
		       __FILE__,__LINE__, unmapstart, size,
		       unmap->buffer);
		unmap->state = ACCEL_WAIT;
		(accel_map.accel->execute)(accel_map.accel, unmap);
	}

	DEBUG2("%s:%i: domap %.8lx size %lx to %.8lx\n",
	       __FILE__,__LINE__, domapstart, size, domap->buffer);

	domap->state = ACCEL_FILL;
	
	if (pingpong == 0)
	{
		pingpong = 1;
		
		/* Ensure that the page usage count is incremented! */
		atomic_inc(&mem_map[MAP_NR(accel_map.ping.buffer)].count);
		
		memset(accel_map.ping.buffer, 0xff, PAGE_SIZE);
		
		return (unsigned long) accel_map.ping.buffer;
	}
	else
	{
		pingpong = 0;
		
		/* Ensure that the page usage count is incremented! */
       		atomic_inc(&mem_map[MAP_NR(accel_map.pong.buffer)].count);
		
		memset(accel_map.pong.buffer, 0xff, PAGE_SIZE);
		
		return (unsigned long) accel_map.pong.buffer;
	}
}

static struct vm_operations_struct graph_accel_vmops = {

	NULL,                   /* open		*/
	NULL,			/* close	*/
	NULL,
	NULL,                   /* protect	*/
	NULL,                   /* sync		*/
	NULL,                   /* advise	*/
	graph_accel_nopage,	/* nopage	*/
	NULL,                   /* wppage	*/
	NULL,                   /* swapout	*/
	NULL,                   /* swapin	*/
};

static int graph_accel_mmap(struct file *file, struct vm_area_struct *vma)
{
	/* We allow userspace to allocate as much virtual memory as
	 * the OS allows, since all we ever have to allocate is two
	 * physical pages.
	 */

	vma->vm_ops = &graph_accel_vmops;

	vma->vm_file = file;
	file->f_count++;
	
	pingpong = 0;
	
	accel_map.file = file;
	
	memset(accel_map.ping.buffer, 0xff, PAGE_SIZE);
	memset(accel_map.pong.buffer, 0xff, PAGE_SIZE);
	 
	accel_map.ping.state = accel_map.pong.state = ACCEL_IDLE;
	
	return 0;
}

static int kgiproc_mmap_gc(struct file *file, struct vm_area_struct *vma)
{
	unsigned long size = vma->vm_end - vma->vm_start;
	
	if (size != PAGE_SIZE) return -EINVAL;
	
	kgifb_graph.gc->version = 0;
	kgifb_graph.gc->oldversion = 0;
	
	vma->vm_ops = &graph_GC_vmops;

	vma->vm_file = file;
	file->f_count++;
	
	return 0;
}

static int kgiproc_mmap_generic(struct file *file, struct vm_area_struct *vma)
{
	unsigned long size = vma->vm_end - vma->vm_start;
	unsigned long offset;
	struct kgi_mmio_region *regptr;

	regptr = kgifb_display.mode.mmio;
	while (strnicmp(file->f_dentry->d_name.name, regptr->name, 
			strlen(file->f_dentry->d_name.name)) != 0)
	  regptr = regptr->next;
	
	offset = regptr->pstart + vma->vm_offset;
	
	if (vma->vm_offset + size > regptr->psize)
	  return -EINVAL;
	pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
	
	if (remap_page_range(vma->vm_start, offset, size, vma->vm_page_prot))
	  return -EAGAIN;
	
	vma->vm_file = file;
	return 0;
}

static struct file_operations kgi_proc_gc_fileops = 
{
	NULL,		     	/* lseek	*/
	NULL,		      	/* read		*/
	NULL,			/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* poll		*/
	NULL,			/* ioctl	*/
	kgiproc_mmap_gc,       	/* mmap		*/
	NULL,			/* open		*/
	NULL,			/* flush	*/
	NULL,			/* release	*/
	NULL			/* fsync	*/
};

static struct inode_operations kgi_proc_gc_inodeops = 
{
	&kgi_proc_gc_fileops,
};

static struct proc_dir_entry kgi_proc_gc_entry = 
{
	0, 
	2, 
	"gc", 
	S_IFREG | S_IRUSR | S_IWUSR | S_IRUGO | S_IWUGO | S_IROTH | S_IWOTH, 
	1, 
	0, 
	0, 
	0, 
	&kgi_proc_gc_inodeops
};

static struct file_operations kgi_proc_kgicommand_fileops = 
{
	NULL,		     	/* lseek	*/
	NULL,		      	/* read		*/
	NULL,			/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* poll		*/
	NULL,			/* ioctl	*/
	graph_accel_mmap,   	/* mmap		*/
	NULL,		      	/* open		*/
	NULL,			/* flush	*/
	NULL,			/* release	*/
	NULL			/* fsync	*/
};

static struct inode_operations kgi_proc_kgicommand_inodeops = 
{
	&kgi_proc_kgicommand_fileops,
};

static struct proc_dir_entry kgi_proc_kgicommand_entry = 
{
	0,
	10, 
	"kgicommand", 
	S_IFREG | S_IRUSR | S_IWUSR | S_IRUGO | S_IWUGO | S_IROTH | S_IWOTH, 
	1, 
	0, 
	0, 
	0, 
	&kgi_proc_kgicommand_inodeops
};

struct kgiproc_driver_entry
{
	struct kgiproc_driver_entry *next;
	struct proc_dir_entry *direntry;
};

static struct kgiproc_driver_entry *kgiproc_driverlist_head = NULL;

/* Allocate struct proc_dir_entry */
static struct proc_dir_entry *alloc_pde(size_t namelen)
{
	struct proc_dir_entry *ret;

	ret = vmalloc(sizeof(struct proc_dir_entry));
	if (!ret) return NULL;

	ret->name = vmalloc(namelen+1); /* Allow \0 terminated strings */
	if (!ret->name) {
		vfree(ret);
		return NULL;
	}
	ret->low_ino = 0;
	ret->namelen = namelen;
	ret->get_info = NULL;
#ifdef THIS_MODULE
	ret->owner = THIS_MODULE;
#else
	ret->fill_inode = NULL;
#endif
	ret->next = NULL;
	ret->parent = NULL;
	ret->subdir = NULL;

	return ret;
}


static void free_pde(struct proc_dir_entry *pde)
{
	if (pde->name) vfree((char*)pde->name);
	vfree(pde);
}


static int kgiproc_init(void)
{
	int retval = 0;
	struct kgiproc_driver_entry *drvptr;
	struct proc_dir_entry   *drvdirent;
	struct inode_operations *drviops;
	struct file_operations  *drvfops;
	struct kgi_mmio_region  *mmioreg;

	the_info.proc_entry = NULL;

	/* Expose all registered drivers */
	/* We only handle one driver for now */

	/* Set up driver root */
	drvptr = vmalloc(sizeof(struct kgiproc_driver_entry));
	if (!drvptr) return -ENOMEM;
	kgiproc_driverlist_head = drvptr;
	
	drvdirent = alloc_pde(7); /* Allows 9999 framebuffers... */
	if (!drvdirent) {
		retval = -ENOMEM;
		goto root_fail;
	}
	drvptr->direntry = drvdirent;
	
	sprintf((char *)drvdirent->name, "gfx%d",
		GET_FB_IDX(the_info.info.node));
	drvdirent->mode = S_IFDIR | S_IRUGO | S_IXUGO;
	drvdirent->nlink = 2;
	drvdirent->uid = 0;
	drvdirent->gid = 0;
	drvdirent->size = 0;
	drvdirent->ops = &proc_dir_inode_operations;

	/* Register driver root */
	retval = proc_register(&proc_root, drvptr->direntry);
	if (retval != 0) {
		goto root_fail;
	}

	/* Register common root entries */
	retval = proc_register(drvptr->direntry, &kgi_proc_gc_entry);
	if (retval != 0) {
		goto root_fail;
	}
	retval = proc_register(drvptr->direntry, &kgi_proc_kgicommand_entry);
	if (retval != 0) {
		goto root_fail;
	}

	/* Map all kgi_mmio_regions exported by the driver */
	mmioreg = kgifb_display.mode.mmio;
	retval = -ENOMEM;
	while (mmioreg != NULL) {
		struct proc_dir_entry *mapdirent;
		struct kgiproc_entry  *thisent;

		thisent = vmalloc(sizeof(struct kgiproc_entry));
		if (!thisent) {
			retval = -ENOMEM;
			goto mmio_fail;
		}

		mapdirent = alloc_pde(strlen(mmioreg->name));
		if (!mapdirent) {
			vfree(thisent);
			retval = -ENOMEM;
			goto mmio_fail;
		}
		strcpy((char *)mapdirent->name, mmioreg->name);
		mapdirent->mode = 
			S_IFREG | S_IRUSR | S_IWUSR | S_IRUGO | S_IWUGO
			| S_IROTH | S_IWOTH;
		mapdirent->nlink = 1;
		mapdirent->uid = 0;
		mapdirent->gid = 0;
		mapdirent->size = 0;

		drviops = vmalloc(sizeof(struct inode_operations));
		if (!drviops) {
			free_pde(mapdirent);
			vfree(thisent);
			retval = -ENOMEM;
			goto mmio_fail;
		}
		mapdirent->ops = drviops;
	
		drviops->create = NULL;
		drviops->lookup = NULL;
		drviops->link = NULL;
		drviops->unlink = NULL;
		drviops->symlink = NULL;
		drviops->mkdir = NULL;
		drviops->mknod = NULL;
		drviops->rename = NULL;
		drviops->readlink = NULL;
		drviops->follow_link = NULL;
		drviops->readpage = NULL;
		drviops->writepage = NULL;
		drviops->bmap = NULL;
		drviops->truncate = NULL;
		drviops->permission = NULL;
		drviops->smap = NULL;
		drviops->updatepage = NULL;
		drviops->revalidate = NULL;

		drvfops = vmalloc(sizeof(struct file_operations));
		if (!drvfops) {
			vfree(drviops);
			free_pde(mapdirent);
			vfree(thisent);
			retval = -ENOMEM;
			goto mmio_fail;
		}
		drviops->default_file_ops = drvfops;
		
		drvfops->llseek = NULL;
		drvfops->read = NULL;
		drvfops->write = NULL;
		drvfops->readdir = NULL;
		drvfops->poll = NULL;
		drvfops->ioctl = NULL;
		drvfops->mmap = &kgiproc_mmap_generic;
		drvfops->open = NULL;
		drvfops->flush = NULL;
		drvfops->release = NULL;
		drvfops->fsync = NULL;
		drvfops->fasync = NULL;
		drvfops->check_media_change = NULL;
		drvfops->revalidate = NULL;
		drvfops->lock = NULL;

		retval = proc_register(drvptr->direntry, mapdirent);
		if (retval != 0) {
			vfree(drvfops);
			vfree(drviops);
			free_pde(mapdirent);
			vfree(thisent);
			goto mmio_fail;
		}

		/* Success - add to beginning of list. */
		thisent->entry = mapdirent;
		thisent->next = the_info.proc_entry;
		the_info.proc_entry = thisent;

		mmioreg = mmioreg->next;
	}

	return 0;

  mmio_fail:
	printk("fbcon-kgi.c: can't clean up after error...\n");
	return retval;

  root_fail:
	kgiproc_exit();

	return retval;
}


static void kgiproc_exit(void)
{
	struct kgiproc_driver_entry *drvptr = kgiproc_driverlist_head;
	struct kgiproc_entry *curr_ent = the_info.proc_entry;

	if (!drvptr) return;

	while (curr_ent) {
		struct kgiproc_entry *free_ent = curr_ent;

		vfree(curr_ent->entry->ops->default_file_ops);
		vfree(curr_ent->entry->ops);
		free_pde(curr_ent->entry);
		curr_ent = curr_ent->next;
		vfree(free_ent);
	}

	if (kgi_proc_kgicommand_entry.low_ino) {
		proc_unregister(drvptr->direntry,
			      kgi_proc_kgicommand_entry.low_ino);
	}
	if (kgi_proc_gc_entry.low_ino) {
		proc_unregister(drvptr->direntry, kgi_proc_gc_entry.low_ino);
	}

	if (drvptr->direntry) {
		if (drvptr->direntry->low_ino) {
			proc_unregister(&proc_root, drvptr->direntry->low_ino);
		}
		free_pde(drvptr->direntry);
	}
	vfree(drvptr);
	
	kgiproc_driverlist_head = NULL;
}


static void
pingpong_exec(struct kgi_accel *accel, struct kgi_accel_buffer *peng)
{
	char *acc = (char *)peng->buffer;
	kgiu32 cmd;
	int type, res = 0;
	int offset = 0;
	
	cmd = *((kgiu32 *)(acc));
	
	acc += 4;
	offset += 4;

	while ((offset < PAGE_SIZE) && (cmd != 0xffffffff))
	{
		if ((kgifb_graph.io_cnt = _IOC_SIZE(cmd)) == -1)
		{
			/* We got problems.... */
			printk("get_accel_size: Unknown accel size!\n");
			peng->state = ACCEL_IDLE;
			return;
		}
		kgifb_graph.io_res = NULL;
		memcpy(kgifb_graph.io_buf, acc, kgifb_graph.io_cnt);
	
		type = _IOC_TYPE(cmd) - KGI_IOCTLBASE;
		res = kgifb_display.cmd[type](&kgifb_graph, cmd, 0);
		
		offset += kgifb_graph.io_cnt;
		acc += kgifb_graph.io_cnt;
		cmd = *((kgiu32 *)(acc));
		acc += 4;
		offset += 4;
	}
	peng->state = ACCEL_IDLE;
}


static void pingpong_wait(struct kgi_accel *accel)
{
}

#endif /* defined(CONFIG_PROC_FS) && defined(KGICON_USE_PROC) */


#ifdef MODULE

EXPORT_NO_SYMBOLS;

int init_module(void)
{
	int err;

	err = kgifb_init();
	if (err) {
		return err;
	}

	return 0;
}

void cleanup_module(void)
{	
	kgifb_exit();
}
#endif
